#pragma rtGlobals=1		// Use modern global access method.
#pragma IgorVersion=6.01		// AfterWindowCreatedHook requires 6.01
#pragma version = 1.1.0.0
#pragma IndependentModule=ACL_WindowDesktops

// ********************
//  LICENSE
// ********************
//	Copyright (c) 2007 by Adam Light
//	
//	Permission is hereby granted, free of charge, to any person
//	obtaining a copy of this software and associated documentation
//	files (the "Software"), to deal in the Software without
//	restriction, including without limitation the rights to use,
//	copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the
//	Software is furnished to do so, subject to the following
//	conditions:
//	
//	The above copyright notice and this permission notice shall be
//	included in all copies or substantial portions of the Software.
//	
//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
//	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
//	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
//	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
//	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//	OTHER DEALINGS IN THE SOFTWARE.
//
// *************************
//  VERSION HISTORY
// *************************
// For a complete list of all changes made to this code, please view the Subversion
// commit message log for this project at:
//   http://www.igorexchange.com/project/cvs/59

// ********************
//  CONSTANTS
// ********************
static StrConstant kPackageName = "ACL_WindowDesktops"
static StrConstant kPreferencesFileName = "ACL_WindowDesktopsPrefs.bin"
static Constant kPreferencesRecordID = 0
static Constant K_CURRENTPREFSVERSION = 100

// ********************
//  STRUCTURES
// ********************
Structure ACL_WindowDesktopsStr
	uint32 version
	double panelCoords[4]		// left, top, right, bottom
	uint32 reserved[100]		// reserved for future use
EndStructure

// ********************
//  MENUS
// ********************
Menu "Misc", hideable
	"Desktop Control Panel",/Q,Initialize_pnlDesktopControl()
End

// ********************
//  PANEL
// ********************
Function BuildpnlDesktopControl()
	PauseUpdate; Silent 1		// building window...
	NewPanel /W=(702,503,945,550)/FLT=1/K=1/N=pnlDesktopControl as "Desktop Control"
	Button buttonAssignDefault,pos={48,2},size={70,20},proc=ButtonAssignDesktop,title="Default"
	Button buttonAssignDefault,userdata(desktop)=  "0"
	Button buttonAssign1,pos={120,2},size={20,20},proc=ButtonAssignDesktop,title="1"
	Button buttonAssign1,userdata(desktop)=  "1"
	Button buttonAssign2,pos={142,2},size={20,20},proc=ButtonAssignDesktop,title="2"
	Button buttonAssign2,userdata(desktop)=  "2"
	Button buttonAssign3,pos={164,2},size={20,20},proc=ButtonAssignDesktop,title="3"
	Button buttonAssign3,userdata(desktop)=  "3"
	Button buttonAssign4,pos={187,2},size={20,20},proc=ButtonAssignDesktop,title="4"
	Button buttonAssign4,userdata(desktop)=  "4"
	Button buttonShowAll,pos={210,25},size={30,20},proc=ButtonShowDesktop,title="All"
	Button buttonShowAll,userdata(desktop)=  "-1"
	Button buttonShow1,pos={120,25},size={20,20},proc=ButtonShowDesktop,title="1"
	Button buttonShow1,userdata(desktop)=  "1"
	Button buttonShow2,pos={142,25},size={20,20},proc=ButtonShowDesktop,title="2"
	Button buttonShow2,userdata(desktop)=  "2"
	Button buttonShow3,pos={164,25},size={20,20},proc=ButtonShowDesktop,title="3"
	Button buttonShow3,userdata(desktop)=  "3"
	Button buttonShow4,pos={187,25},size={20,20},proc=ButtonShowDesktop,title="4"
	Button buttonShow4,userdata(desktop)=  "4"
	Button buttonShowDefault,pos={48,25},size={70,20},proc=ButtonShowDesktop,title="Default"
	Button buttonShowDefault,userdata(desktop)=  "0"
	TitleBox titleAssign,pos={4,6},size={41,13},title="Assign:",frame=0,fStyle=1
	TitleBox titleShow,pos={4,29},size={35,13},title="Show:",frame=0,fStyle=1
	Button buttonHelp,pos={220,3},size={20,20},proc=ButtonHelp,title="?",fSize=14
	Button buttonHelp,fStyle=1
	SetActiveSubwindow _endfloat_
EndMacro

// ********************
//  INITIALIZATION
// ********************
Function Initialize_pnlDesktopControl()
	String panelName = "pnlDesktopControl"
	DoWindow $(panelName)
	if (!V_flag)
		String curDataFolder = GetDataFolder(1)
		if (!DataFolderExists("root:Packages:ACL_WindowDesktops"))
			NewDataFolder/O/S root:Packages
			NewDataFolder/O/S root:Packages:ACL_WindowDesktops
		else
			SetDataFolder root:Packages:ACL_WindowDesktops
		endif
		
		Variable displayedDesktop = GetDisplayedDesktop()
		if (numtype(displayedDesktop) != 0)
			// Create the variable and set to display the default desktop.
			SetDisplayedDesktop(0)
			displayedDesktop = GetDisplayedDesktop()
		endif
		
		BuildpnlDesktopControl()
		SetWindow pnlDesktopControl hook(ACL_DesktopControlHook)=ACL_DesktopControlHook
		STRUCT ACL_WindowDesktopsStr prefs
		LoadDesktopControlPrefs(prefs)
		MoveWindow/W=pnlDesktopControl prefs.panelCoords[0], prefs.panelCoords[1], prefs.panelCoords[2], prefs.panelCoords[3] 	
		Variable errorNum = ChangeDisplayedDesktop(displayedDesktop)
		if (errorNum != 0)
			printf "Error %d in Initialize_pnlDesktopControl.\r", errorNum
		endif	
		SetDataFolder curDataFolder	
	endif
	
	// Register a function provided by this module that responds to creation of any window
	// and assigns that window to the currently selected desktop.
	RegisterWindowCreatedHook()
End


// ********************
//  ACTION FUNCTIONS
// ********************
Function ButtonAssignDesktop(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			// click code here
			Variable desktopNum = str2num(GetUserData(ba.win, ba.ctrlName, "desktop"))
			String topWindowName = WinName(0, 4183)	// include all window types EXCEPT procedure windows, since they don't have names
			Variable errorNum = SetWindowDesktop(topWindowName, desktopNum)
			if (errorNum != 0)
				printf "Error %d in ButtonAssignDesktop().\r", errorNum
			endif

			break
	endswitch

	return 0
End

Function ButtonShowDesktop(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			Variable desktopNum = str2num(GetUserData(ba.win, ba.ctrlName, "desktop"))
			Variable errorNum = ChangeDisplayedDesktop(desktopNum)
			if (errorNum != 0)
				printf "Error %d in ButtonShowDesktop().\r", errorNum
			endif
			break
	endswitch

	return 0
End

Function ButtonHelp(ba) : ButtonControl
	STRUCT WMButtonAction &ba

	switch( ba.eventCode )
		case 2: // mouse up
			DisplayHelpTopic/Z "Window Desktops"
			if (V_Flag != 0)	// help topic not found
				DoAlert 0, "The ACL_WindowDesktops.ihf help file could not be found.  It should be in the same directory as the ACL_WindowDesktops.ipf procedure file."
			endif
			break
	endswitch

	return 0
End

// ********************
//  AUXILLARY FUNCTIONS
// ********************
//**
// Determine what virtual desktop a window is currently assigned to.
//
// @param windowName
// 	Name of window to check.
//
// @return
// 	The number of the desktop that windowName is assigned
// 	to or -1 if the window does not exist.
//*
Function GetWindowDesktop(windowName)
	String windowName
	Variable desktopNum
	
	// check to see if window exists
	Variable type = WinType(windowName)
	if (type && type != 13)		// exclude XOP target windows
		desktopNum = str2num(GetUserData(windowName, "", "ACL_desktopNum"))
		if (numtype(desktopNum) != 0)		// value is undefined
			desktopNum = 0		// show on all desktops
		else
			desktopNum = round(desktopNum)
		endif
	else
		desktopNum = -1		// window doesn't exist
	endif
	
	return desktopNum
End

//**
// Assign a window to a virtual desktop.
//
// @param windowName
// 	Name of window.
// @param desktopNum
// 	Number of desktop to assign window to.  If 0, window will be
// 	displayed on all desktops.  If >= 1, window will be displayed
// 	only on a specific desktop.
//
// @return
// 	Zero if the function is completed successfully or a negative
// 	value indicating an error code.
//*
Function SetWindowDesktop(windowName, desktopNum)
	String windowName
	Variable desktopNum
	Variable errorValue
	
	// make sure desktopNum parameter is valid
	if (numtype(desktopNum) == 0)		// value is a real number
		if (desktopNum >= 0)
			desktopNum = round(desktopNum)		// desktopNum must be an integer
		endif
	else
		errorValue = -1
		return errorValue
	endif
	
	// check to see if window exists
	Variable type = WinType(windowName)
	if (type)	
		if (type == 13)	// exclude XOP target windows
			printf "Can't assign a target window of an XOP to a desktop.\r"
			return 0
		else
			SetWindow $(windowName) userdata(ACL_desktopNum)=num2istr(desktopNum)
		endif
	else
		errorValue = -1		// window doesn't exist
	endif	
	
	if (errorValue == 0)		// If there is no error, call ChangeDisplayedDesktop so the window just set is hidden or visible as appropriate.
		UpdateWindowShowHide(windowName)
	endif
	
	return errorValue		// 0 if successful
End

//**
// Update a window's show/hide status as approrpiate.
//
// @param windowName
//*
Function UpdateWindowShowHide(windowName)
	String windowName
	
	Variable currentDesktop = GetDisplayedDesktop()
	Variable windowsDesktop = GetWindowDesktop(windowName)
	Variable hide
	if (numtype(currentDesktop) == 0 && windowsDesktop >= 0)	// no error
		if (currentDesktop == -1)		// show ALL windows
			hide = 0
		elseif (windowsDesktop == 0 || windowsDesktop == currentDesktop)
			hide = 2		// show this window
		else
			hide = 1		// hide this window
		endif
		SetWindow $(windowName) hide=hide, needUpdate=1
	endif
End

//**
// Get the number of the desktop that is currently displayed.
//
// @return
// 	A number indicating the currently displayed desktop.
// 	If the currentDisplayedDesktop variable does not exist,
// 	NaN will be returned.
//*
Function GetDisplayedDesktop()
	NVAR/Z currentDisplayedDesktop = root:Packages:ACL_WindowDesktops:currentDisplayedDesktop
	if (NVAR_Exists(currentDisplayedDesktop))
		return currentDisplayedDesktop		
	else
		return NaN
	endif
End

//**
// Set the number of the currently displayed desktop.
// Instead of changing the value directly, other functions should
// call this function to set the value.
//
// @param desktopNum
// 	The value of the currently displayed desktop.  A value
// 	of -1 means "All", 0 means "Default", and a positive
//	integer represents that desktop.
//*
Function SetDisplayedDesktop(desktopNum)
	Variable desktopNum
	
	NVAR/Z currentDisplayedDesktop = root:Packages:ACL_WindowDesktops:currentDisplayedDesktop
	if (!NVAR_Exists(currentDisplayedDesktop))
		// Make sure the data folder where this variable will go exists.
		NewDataFolder/O root:Packages
		NewDataFolder/O root:Packages:ACL_WindowDesktops
		Variable/G root:Packages:ACL_WindowDesktops:currentDisplayedDesktop
		NVAR currentDisplayedDesktop = root:Packages:ACL_WindowDesktops:currentDisplayedDesktop
	endif
	currentDisplayedDesktop = desktopNum
End

//**
// Chang the virtual desktop that is currently displayed.
//
// @param newDesktopNum
// 	Number of desktop to switch to (if >= 1).  Additionally,
// 	there are a few special cases for this parameter:
// 	0	Show all windows that are present on any desktop, including
//		windows that have not yet been assigned to a desktop.
// 	-1	Show all windows, regardless of the desktop they are assigned to.
// @param forceUpdate
// 	[OPTIONAL]  If set to 1, the desktop will be "changed" even if
// 	newDesktopNum is the same as the current desktop.  This can be
// 	used to force a refresh of the windows currently being displayed.
//
// @return
// 	Zero if there was no error or a negative number representing the error code.
//*
Function ChangeDisplayedDesktop(newDesktopNum, [forceUpdate])
	Variable newDesktopNum
	Variable forceUpdate		// set to 1 if the desktop should be changed even if currentDesktop = newDesktopNum

	if (ParamIsDefault(forceUpdate))
		forceUpdate = 0
	endif
	
	// Only force an update of the windows if the displayed desktop is changing.
	Variable currentDesktop = GetDisplayedDesktop()
	if ((numtype(currentDesktop) != 0) || (currentDesktop != newDesktopNum))
		forceUpdate = 1
	endif
	
	Variable errorValue
	// make sure newDesktopNum is valid
	if (numtype(newDesktopNum) != 0)		// newDesktopNum is invalid
		errorValue = -1
		return errorValue		// error: invalid newDesktopNum
	elseif (newDesktopNum < -1)
		errorValue = -1
		return errorValue		// error: invalid newDesktopNum
	else
		newDesktopNum = round(newDesktopNum)
	endif

	// Sets the global variable that keeps track of which desktop
	// is currently displayed.
	SetDisplayedDesktop(newDesktopNum)

	if (forceUpdate == 1)
		// get a list of all windows that GetWindow/SetWindow can be used with
		String windowList = WinList("*", ";", "WIN:87")	// include all window types EXCEPT procedure windows, since they don't have names
		Variable numWindows = ItemsInList(windowList, ";")
		Variable n, winDesktopNum, hide
		String currentWinName
		For (n=0; n<numWindows; n+=1)
			currentWinName = StringFromList(n, windowList, ";")
			UpdateWindowShowHide(currentWinName)
		EndFor
	Endif
	
	if (errorValue == 0)
		SetShowButtonColors(newDesktopNum, "pnlDesktopControl")
		
		// If the Wavemetrics GraphBrowserPanel is visible, then force it to update
		// so that if the user has unchecked the "Are invisible" checkbox on that 
		// panel the graphs displayed on the panel will be updated when desktops are changed.
		//
		// This is sort of a hack, because independent modules cannot call functions
		// in other independent modules other than by using Execute.  Ideally we'd just
		// create a WMWinHookStruct here, set the window name and event code, and
		// pass that to the GraphBrowser's window hook function.  Alas, global structures
		// do not (yet) exist in Igor Pro, so that won't work either.  We at least check
		// that the target function exists and that it takes the parameter we're prepared
		// to use in the call to the function in an attempt to avoid errors if WM changes
		// either of the functions GrfBrowserPanelWinProc() or UpdateGraphList().
		if (WinType("GraphBrowserPanel") == 7)		// The window is a panel.
			String updateFxInfo = FunctionInfo("WM_GrfBrowser#UpdateGraphList")
			if (strlen(updateFxInfo) > 0)
				Variable numParams = NumberByKey("N_PARAMS", updateFxInfo, ":", ";")
				Variable param0Type = NumberByKey("PARAM_0_TYPE", updateFxInfo, ":", ";")
				if (numParams == 1 && param0Type == 8192)		// param0Type is a string
					Execute "WM_GrfBrowser#UpdateGraphList(\"GraphBrowserPanel\")"
				endif
			endif
		endif
	endif	
	return errorValue
End

//**
// Set the colors of the buttons on the Desktop Control panel.
//
// @param desktopNum
// 	The number of the desktop that is currently displayed.
// @param windowName
// 	The name of the Desktop Control panel.
//*
Function SetShowButtonColors(desktopNum, windowName)
	Variable desktopNum
	String windowName
	
	Variable errorNum
	DoWindow pnlDesktopControl
	if (V_Flag)
		String buttonList, currentCtrlName
		Variable n, numButtons
		Variable desktopControlled		// the number of the desktop controlled by the button
		buttonList = ControlNameList(windowName, ";", "ButtonShow*")
		numButtons = ItemsInList(buttonList, ";")
		For (n=0; n<numButtons; n+=1)
			currentCtrlName = StringFromList(n, buttonList, ";")
			ControlInfo/W=$(windowName) $(currentCtrlName)
			if (abs(V_flag) == 1)	// control is a button
				desktopControlled = str2num(GetUserData(windowName, currentCtrlName, "desktop"))
				if (numtype(desktopControlled) == 0)
					if (desktopControlled == desktopNum)
						Button $(currentCtrlName) win=$(windowName), fColor=(0,43520,65280)		// blue button
					else
						Button $(currentCtrlName) win=$(windowName), fColor=(0,0,0)				// normal colored button
					endif
				endif
			endif			
		EndFor
	endif
End

//**
// Set a hook function for the Igor AfterWindowCreatedHook hook type.
//*
Function RegisterWindowCreatedHook()
	SetIgorHook AfterWindowCreatedHook = Hook_WindowCreated
End
 
// ********************
//  HOOK FUNCTIONS
// ********************
Function ACL_DesktopControlHook(str)
	STRUCT WMWinHookStruct &str
	Variable statusCode = 0
	Switch (str.eventCode)
		Case 12:	// window was moved: store new coordinates
			STRUCT ACL_WindowDesktopsStr prefs
			SaveDesktopControlPrefs(prefs)
			break
		default:			
	EndSwitch
	return statusCode
End

Function Hook_WindowCreated(windowNameStr, type)
	String windowNameStr
	Variable type

	Variable currentDesktop = GetDisplayedDesktop()
	if (numtype(currentDesktop) == 0 && currentDesktop > 0)
		SetWindowDesktop(windowNameStr, currentDesktop)
	endif
	return 0
End

// ********************
//  PACKAGE PREFS LOADING/SAVING
// ********************
Function LoadDesktopControlPrefs(s)
	STRUCT ACL_WindowDesktopsStr &s
	
	// This loads preferences from disk if they exist on disk.
	LoadPackagePreferences kPackageName, kPreferencesFileName, kPreferencesRecordID, s

	// If error, prefs not found or not compatible, initialize them.
	if (V_flag!=0 || V_bytesRead==0 || s.version!=K_CURRENTPREFSVERSION)
		SaveDesktopControlPrefs(s)	// Create default prefs file.
	endif
	
	// code that loads prefs

End

Function SaveDesktopControlPrefs(s)
	STRUCT ACL_WindowDesktopsStr &s
	
	// fill in structure with current values
	s.version = K_CURRENTPREFSVERSION
	GetWindow pnlDesktopControl wsize			// get current window coordinates
	s.panelCoords[0] = V_left		// Left
	s.panelCoords[1] = V_top		// Top
	s.panelCoords[2] = V_right		// Right
	s.panelCoords[3] = V_bottom		// Bottom
	
	Variable i
	for(i=0; i<100; i+=1)
		s.reserved[i] = 0
	endfor

	SavePackagePreferences kPackageName, kPreferencesFileName, kPreferencesRecordID, s
	return 0	
End